#include <Adafruit_BMP085.h>
#include <NTPClient.h>
#include <Adafruit_GFX.h>
#include <gfxfont.h>
#include <Fonts/FreeSansBold9pt7b.h>
#include <Fonts/FreeSans9pt7b.h>
#include "Adafruit_SSD1306.h"
#include <WEMOS_SHT3X.h>
#include <ESP8266WiFi.h>
#include <WiFiUdp.h>
#include <ESP8266HTTPClient.h>
#include <Losant.h>

#define SHT30
//#define BMP085

// WiFi credentials.
const char* WIFI_SSID = "your-wifi-network-name";
const char* WIFI_PASS = "your-wifi-password";

// Losant credentials.
const char* LOSANT_DEVICE_ID = "your-losant-device-id-1"; // upstairs node
//const char* LOSANT_DEVICE_ID = "your-losant-device-id-2"; // study node
//const char* LOSANT_DEVICE_ID = "your-losant-device-id-3"; // downstairs node
//const char* LOSANT_DEVICE_ID = "your-losant-device-id-4"; // outside node
const char* LOSANT_ACCESS_KEY = "your-losant-access-key";
const char* LOSANT_ACCESS_SECRET = "your-losant-access-secret";

#define OLED_RESET 0  // GPIO0
Adafruit_SSD1306 display(OLED_RESET);

#ifdef SHT30
SHT3X sht30(0x45);
#endif
#ifdef BMP085
Adafruit_BMP085 bmp085;
#endif

// Create a secure WiFi client. This can also be an instance of the unsecured
// WiFiClient class, but the secure TLS client is recommended.
WiFiClientSecure wifiClient;
WiFiUDP ntpUDP;

NTPClient timeClient(ntpUDP, "pool.ntp.org", -7*3600, 60000); //-7 for PDT, -8 for PST

// Create an instance of a Losant device.
LosantDevice device(LOSANT_DEVICE_ID);

// Connects to WiFi and then to the Losant platform.
void connectLosant() 
{
  int startTime = millis();
  
  display.clearDisplay();
  display.setFont();
  display.setCursor(0,0);
  display.println("connect...");
  //display.display();
  
  // Connect to WiFi.
  Serial.print("Connecting to ");
  Serial.println(WIFI_SSID);

  display.setCursor(0,20);
  display.println("WiFi: ");
  display.display();
  
  if (WiFi.status() != WL_CONNECTED) 
  {
    // WiFi fix: https://github.com/esp8266/Arduino/issues/2186
    WiFi.persistent(false);
    WiFi.mode(WIFI_OFF);
    WiFi.mode(WIFI_STA);
    WiFi.begin(WIFI_SSID, WIFI_PASS);
    
    while (WiFi.status() != WL_CONNECTED && (millis() - startTime < 15000)) 
    {
      if (WiFi.status() == WL_CONNECT_FAILED) 
      {
        Serial.println("Failed to connect to WIFI. Please verify credentials.");
        return;
      }
        
      delay(500);
      Serial.print(".");
    }
  }

  if (WiFi.status() != WL_CONNECTED)
    return;
    
  display.setCursor(50,20);
  display.println("OK");
  display.setCursor(0,30);
  display.println("Losant: ");
  display.display();
  
  Serial.println();
  Serial.println("WiFi connected");
  Serial.println("IP address: ");
  Serial.println(WiFi.localIP());

  // Connect to Losant.
  Serial.println();
  Serial.print("Connecting to Losant...");

  // Connect the device instance to Losant using TLS encryption.
  device.connectSecure(wifiClient, LOSANT_ACCESS_KEY, LOSANT_ACCESS_SECRET);

  while(!device.connected() && (millis() - startTime < 15000)) 
  {
    delay(500);
    Serial.print(".");
  }

  if (!device.connected())
  {
    display.setCursor(0,40);
    display.println("failed");
    display.display();
    WiFi.mode(WIFI_OFF);
    delay(5000);
    return;
  }
    
  display.setCursor(50,30);
  display.println("OK");
  display.display();
  
  Serial.println("Connected!");
}

// Command callback function. Invoked whenever this device receives
// a command from the Losant platform.
void handleCommand(LosantCommand *command) 
{
  Serial.print("Command received: ");
  Serial.println(command->name);

  // Optional command payload. May not be present on all commands.
  JsonObject& payload = *command->payload;

  // Perform action specific to the command received.
  if(strcmp(command->name, "start recording") == 0) 
  {
    int resolution = payload["resolution"];
  }
}

void updateSensorState()
{
#ifdef SHT30  
  sht30.get(); 

  float dewpointC = 243.04*(log(sht30.humidity/100)+((17.625*sht30.cTemp)/(243.04+sht30.cTemp))) / 
                    (17.625-log(sht30.humidity/100)-((17.625*sht30.cTemp)/(243.04+sht30.cTemp))); 
  float dewpointF = (dewpointC * 9 / 5) + 32;

  float temperatureF = sht30.fTemp;
#endif
#ifdef BMP085
  float temperatureF = bmp085.readTemperature() * 9 / 5 + 32;
  float myAltitudeInM = 100.0f;
  float pressure = (float)bmp085.readSealevelPressure(myAltitudeInM) / 3386.389f; // convert Pascals to in-Hg
#endif

  display.clearDisplay();
  display.setFont();
  display.setCursor(0,0);
  display.println();
  
  display.setFont(&FreeSansBold9pt7b);
  display.setCursor(12,14);
  display.print(temperatureF, 1);

  display.setFont();
  display.setCursor(49,0);
  display.print("o");
  //display.setFont(&FreeSansBold9pt7b);
  //display.setCursor(52,14);
  //display.print("F");

  int line2y = 27;
  display.setFont();
  display.setCursor(0,line2y);
#ifdef SHT30
  display.print("DP");

  display.setFont(&FreeSans9pt7b);
  display.setCursor(17,line2y+6);
  display.print(dewpointF, 1);

  display.setFont();
  display.println();
  display.setCursor(54,line2y-8);
  display.print("o");
#endif
#ifdef BMP085
  display.setFont(&FreeSans9pt7b);
  display.setCursor(0,line2y+6);
  display.print(pressure, 2);

  display.setFont();
  display.println();
  display.setCursor(50,line2y);
  display.print("in");
#endif

  //display.setFont(&FreeSansBold9pt7b);
  //display.setCursor(52,40);
  //display.print("F");

  display.setCursor(0,41);
  //display.print("Fri 10:33A");
  char dateTimeBuf[16];
  char* dayName;
  switch (timeClient.getDay())
  {
    case 0:
      dayName = "Sun";
      break;

    case 1:
      dayName = "Mon";
      break;

    case 2:
      dayName = "Tue";
      break;

    case 3:
      dayName = "Wed";
      break;

    case 4:
      dayName = "Thu";
      break;

    case 5:
      dayName = "Fri";
      break;

    default:
      dayName = "Sat";
      break;
  }

  int hour = timeClient.getHours();
  int hour12 = hour == 0 ? 12 : hour > 12 ? hour - 12 : hour;
  sprintf(dateTimeBuf, "%s %s%d:%02d%c", dayName, hour12 < 10 ? " " : "", hour12, timeClient.getMinutes(), hour < 12 ? 'A' : 'P');
  display.print(dateTimeBuf);
  display.display();

  // create a buffer to hold the state report data
  StaticJsonBuffer<250> jsonBuffer;
  JsonObject& state = jsonBuffer.createObject();
  state["temperature"] = temperatureF;
#ifdef SHT30  
  state["humidity"] = sht30.humidity;
  state["dewpoint"] = dewpointF;
#endif  
#ifdef BMP085
  state["pressure"] = pressure;
#endif

  // Report the state to Losant.
  device.sendState(state);
}

void setup() 
{
  Serial.begin(115200);
  //while(!Serial) { }

  display.begin(SSD1306_SWITCHCAPVCC, 0x3C);  // initialize with the I2C addr 0x3C (for the 64x48)
  display.setTextSize(1);
  //display.setFont(&FreeSansBold9pt7b);
  display.setTextColor(WHITE);
  display.clearDisplay();
  display.setCursor(0,0);
  display.print("init");
  display.display();

#ifdef BMP085
  bmp085.begin();
#endif

  // Pass the command handler function to the device.
  // Only one command handler can be specified.
  device.onCommand(&handleCommand);

  connectLosant();
  
  timeClient.begin();
  timeClient.update();
}

int lastUpdateTime = 0;
bool first = true;

void loop() 
{
  bool toReconnect = false;

  if (WiFi.status() != WL_CONNECTED) 
  {
    Serial.println("Disconnected from WiFi");
    toReconnect = true;
  }

  if (!device.connected()) 
  {
    Serial.println("Disconnected from Losant");
    toReconnect = true;
  }

  if (toReconnect) 
    connectLosant();

  if (device.connected())
  {
    device.loop();
  
    // Every 15 seconds.
    if (first || (millis() - lastUpdateTime > 15000)) 
    {
      updateSensorState();
      lastUpdateTime = millis();
      first = false;
    }
  }
}
